/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */


#include <stdio.h>
#include <stdlib.h>

#include "config_xor.h"

#if defined(WITH_GPG) || defined(WITH_PGP)

#include <unistd.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <sys/wait.h>

#include <string.h>
#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif


#include "samhain.h"
#include "sh_utils.h"
#include "sh_error.h"
#include "sh_tiger.h"

static struct {
  char     conf_id[SH_MINIBUF+1];
  char     conf_fp[SH_MINIBUF+1];
  char     data_id[SH_MINIBUF+1];
  char     data_fp[SH_MINIBUF+1];
} gp;

typedef struct {
  pid_t    pid;
  FILE   * pipe;
} sh_gpg_popen_t;

#define SH_GPG_OK      0
#define SH_GPG_BAD     1
#define SH_GPG_BADSIGN 2


static FILE * sh_gpg_popen (sh_gpg_popen_t  *source, char * filepath, 
			    int mode, char * id)
{
  int pipedes[2];
  FILE * outf;
  char * envp[2];
  char   path[256];
  char   cc1[32];
  char   cc2[32];
#ifdef WITH_PGP
  char   cc3[32];
#endif

  char * arg[6];


#if defined(WITH_GPG)
  /* -- GnuPG -- */
  sl_strlcpy (path,  DEFAULT_GPG_PATH, 256);
  sl_strlcpy (cc1,   _("--status-fd"),  32);
  sl_strlcpy (cc2,   _("--verify"),     32);

  arg[0] = path; 
  arg[1] = cc1;
  arg[2] = "1";
  arg[3] = cc2;
  arg[4] = filepath;
  arg[5] = NULL;
#elif defined(WITH_PGP)
  /* -- PGP -- */
  sl_strlcpy (path,  DEFAULT_PGP_PATH, 256);
  if (mode == 0)
    {
      sl_strlcpy (cc1,   _("+language=en"),  32);
      sl_strlcpy (cc2,   _("-o"),     32);
      sl_strlcpy (cc3,   _("/dev/null"),     32);
      
      arg[0] = path; 
      arg[1] = cc1;
      arg[2] = filepath; 
      arg[3] = cc2; 
      arg[4] = cc3;
      arg[5] = NULL;
      /*
      fprintf (stderr, "ARGV: %s %s %s %s %s %s\n", 
	       arg[0], arg[1], arg[2], arg[3], arg[4], arg[5]); 
      */
     }
  else
    {
      sl_strlcpy (cc1,   _("+language=en"),  32);
      sl_strlcpy (cc2,   _("-kvc"),     32);       
      
      arg[0] = path; 
      arg[1] = cc1;
      arg[2] = cc2;
      arg[3] = id;
      arg[4] = NULL;
      arg[5] = NULL;
    }
#endif

  /* use homedir of effective user
   */
  if (sh.ehome != NULL)
    {
      envp[0] = malloc (sl_strlen(sh.ehome) + 6);
      sprintf (envp[0], "HOME=%s", sh.ehome);           /* known to fit  */
      envp[1] = NULL;
    }
  else
    {
      envp[0] = NULL;
    }

  /* Create the pipe 
   */
  if (pipe(pipedes) < 0) {
    if (envp[0] != NULL) 
      free(envp[0]);
    return (NULL);
  }
  
  source->pid = fork();
  
  /* Failure
   */
  if (source->pid == (pid_t) - 1) 
    {
      close(pipedes[0]);
      close(pipedes[1]);
      if (envp[0] != NULL) 
	free(envp[0]);
      return (NULL);
    }

  if (source->pid == (pid_t) 0) 
    {

      /* child - make read side of the pipe stdout 
       */
      if (dup2(pipedes[STDOUT_FILENO], STDOUT_FILENO) < 0)
	_exit(EXIT_FAILURE);
      
      /* close the pipe descriptors 
       */
      close (pipedes[STDIN_FILENO]);
      close (pipedes[STDOUT_FILENO]);
      
      /* don't leak file descriptors
       */
      sh_unix_closeall (3);

      freopen(_("/dev/null"), "r+", stderr); 
      
      /* We should become privileged if SUID,
       * to be able to read the keyring.
       * We have checked that gpg is OK,
       * AND that only a trusted user could overwrite
       * gpg.
       */
      memset (skey, '\0', sizeof(sh_key_t));
      sl_set_suid(); 
      setuid(geteuid());
      
      /* exec the program */
#if defined(WITH_GPG)
      execve (DEFAULT_GPG_PATH, arg, envp);
#elif defined(WITH_PGP)
      execve (DEFAULT_PGP_PATH, arg, envp);
#endif
      
      /* failed 
       */
      _exit(EXIT_FAILURE);
    }

  /* parent
   */

  if (envp[0] != NULL) 
    free(envp[0]);

  close (pipedes[STDOUT_FILENO]);
  fcntl (pipedes[STDIN_FILENO], F_SETFD, FD_CLOEXEC);

  outf = fdopen (pipedes[STDIN_FILENO], "r");

  if (outf == NULL) 
    {
      kill (source->pid, SIGKILL);
      close (pipedes[STDOUT_FILENO]);
      waitpid (source->pid, NULL, 0);
      source->pid = 0;
      return (NULL);
    }
  
  return (outf);
}


static int sh_gpg_pclose (sh_gpg_popen_t *source)
{
  int status = 0;
  
  if (fclose(source->pipe))
    return (-1);
  
  if (waitpid(source->pid, NULL, 0) != source->pid)
    status = -1;
  
  source->pipe = NULL;
  source->pid = 0;
  return (status);
}
 
/***************
static void sh_gpg_trafo (char * out, char * in, char * filename)
{
  int len;

  sl_strlcpy(out, filename, 256);
  sl_strlcat(out,       ": ", 256+2);
  len = sl_strlen(out);
  strncat(out,         in,  8); 
  strncat(out,        " ",  1);
  strncat(out,     &in[8], 17); 
  strncat(out,       "  ",  2);
  strncat(out,    &in[25], 17);
  strncat(out,        " ",  1);
  strncat(out,    &in[42],  8);
  out[len+KEY_LEN+4] = '\0';
}
***************/

static
int sh_gpg_check_file_sign(char * filepath, char * sign_id, char * sign_fp)
{
  struct stat buf;
  char line[256];
  sh_gpg_popen_t  source;
  int have_id = BAD, have_fp = BAD;
#ifdef WITH_PGP
  char *ptr;
#endif

#ifdef HAVE_GPG_CHECKSUM
  char * test_gpg;
  int  ext_len;
  /*
  char test_gpg_ext[ext_len];
  */
#endif

  /* check whether GnuPG exists and has the correct checksum
   */
#if defined(WITH_GPG)

  if (0 != lstat(DEFAULT_GPG_PATH, &buf))
    return SH_GPG_BAD;
  if (0 != tf_trust_check (DEFAULT_GPG_PATH, SL_YESPRIV))
    return SH_GPG_BAD;
#ifdef HAVE_GPG_CHECKSUM
  ext_len = sl_strlen(DEFAULT_GPG_PATH) + KEY_LEN + 2 + 6 + 1;
  test_gpg = sh_tiger_hash_gpg (DEFAULT_GPG_PATH, TIGER_FILE, 0);
  if (0 != sl_strncmp(GPG_HASH, test_gpg, ext_len))
    {
      SH_FREE(test_gpg);
      return SH_GPG_BAD;
    }
  SH_FREE(test_gpg);
#endif

#elif defined(WITH_PGP)

  if (0 != lstat(DEFAULT_PGP_PATH, &buf))
    return SH_GPG_BAD;
  if (0 != tf_trust_check (DEFAULT_PGP_PATH, SL_YESPRIV))
    return SH_GPG_BAD;
#ifdef HAVE_GPG_CHECKSUM
  ext_len = sl_strlen(DEFAULT_PGP_PATH) + KEY_LEN + 2 + 6 + 1;
  test_gpg = sh_tiger_hash_gpg (DEFAULT_PGP_PATH, TIGER_FILE, 0);
  if (0 != sl_strncmp(GPG_HASH, test_gpg, ext_len))
    {
      SH_FREE(test_gpg);
      return SH_GPG_BAD;
    }
  SH_FREE(test_gpg);
#endif

#endif

  source.pipe   = sh_gpg_popen  ( &source, filepath, 0, NULL );

  if (NULL == source.pipe)
    return SH_GPG_BAD;

  while (NULL != fgets(line, 255, source.pipe))
    {
      if (sl_strlen(line) < 18) 
	continue;
#if defined(WITH_GPG)
      if (0 == sl_strncmp(_("GOODSIG"), &line[9], 7))
	{
	  sl_strlcpy (sign_id, &line[25], SH_MINIBUF+1);
	  sign_id[sl_strlen(sign_id)-1] = '\0';  /* remove trailing '"' */
	  have_id = GOOD;
	} 
      if (0 == sl_strncmp(_("VALIDSIG"), &line[9], 8))
	{
	  strncpy (sign_fp, &line[18], 40);
	  sign_fp[40] = '\0';
	  have_fp = GOOD;
	}
#elif defined(WITH_PGP)
      if (0 == sl_strncmp(_("Good signature"), line, 14))
	{
	  ptr = strchr ( line, '"');
	  ++ptr;
	  sl_strlcpy (sign_id, ptr, SH_MINIBUF+1);
	  sign_id[sl_strlen(sign_id)-1] = '\0'; /* remove trailing dot */
	  sign_id[sl_strlen(sign_id)-2] = '\0'; /* remove trailing '"' */
	  have_id = GOOD;
	}
#endif
    }  
  sh_gpg_pclose (&source);

#ifdef WITH_PGP
  /* get the fingerprint */
  source.pipe   = sh_gpg_popen  ( &source, filepath, 1,  sign_id);
  if (NULL == source.pipe)
    return SH_GPG_BAD;
  while (NULL != fgets(line, 255, source.pipe))
    {
      if (sl_strlen(line) < 18) 
	continue;
      ptr = strtok (line, " ");
      while (ptr)
	{
	  ptr = strtok (NULL, " ");
	  if (ptr && 0 == sl_strncmp (ptr, _("fingerprint"), 11))
	    {
	      ptr = strtok (NULL, " "); /* to '=' */
	      sign_fp[0] = '\0';
	      while (ptr)
		{
		  ptr = strtok (NULL, " "); /* part of fingerprint */
		  sl_strlcat (sign_fp, ptr, SH_MINIBUF+1);
		}
	      sign_fp[sl_strlen(sign_fp)-1] = '\0'; /* remove trailing '\n' */
	      if (sl_strlen(sign_fp) > 0) 
		have_fp = GOOD;
	      break;
	    } 
	} 
    }
  sh_gpg_pclose (&source);
#endif


  if (have_id == GOOD && have_fp == GOOD)
    return SH_GPG_OK;
  else
    return SH_GPG_BADSIGN;
}

int sh_gpg_check_sign (int both)
{
  int status;

  char name[256];
  char * dd, * cc;

  cc = file_path('C', 'R');
  dd = file_path('D', 'R');

  if (cc == NULL || dd == NULL)
    {
      sh_error_handle (SH_ERR_FATAL,
		       _("No configuration file\n"));
      return (-1);
    }

  if (cc[0] != '/' || dd[0] != '/')
    {
      sh_error_handle (SH_ERR_FATAL,
		       _("STARTUP yule    -- user %s\n"),
		       sh.user);
      return (0);
    }


  sl_strlcpy(name, file_path('C', 'R'), 256);
  status = sh_gpg_check_file_sign (name, gp.conf_id, gp.conf_fp);

  sh_error_handle (SH_ERR_ALL,
		   _("CONF SIGUSR: |%s|\n"), gp.conf_id);
  sh_error_handle (SH_ERR_ALL,
		   _("CONF SIGFP:  |%s|\n"), gp.conf_fp);



  if (SH_GPG_OK == status && both == 2)
    {
      sl_strlcpy(name, file_path('D', 'R'), 256);
      status = sh_gpg_check_file_sign (name, gp.data_id, gp.data_fp);
      sh_error_handle (SH_ERR_ALL,
		       _("DATA SIGUSR: |%s|\n"), gp.conf_id);
      sh_error_handle (SH_ERR_ALL,
		       _("DATA SIGFP:  |%s|\n"), gp.conf_fp);
    }
      
  if (SH_GPG_OK == status)
    {
#ifdef USE_FINGERPRINT
      if ((sl_strcmp(SH_GPG_FP, gp.conf_fp) == 0))
	{
	  if ( (both == 2) && (sl_strcmp(gp.data_fp, gp.conf_fp) != 0))
	    {
	      status = SH_GPG_BADSIGN;
	    }
	  else
	    {
	      sh_error_handle (SH_ERR_FATAL,
#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
			       _("STARTUP yule    -- user %s\n"\
				 "         GOODSIG %s  %s\n"\
				 "         FP_KEY  %s"),
#else
			       _("STARTUP samhain -- user %s\n"\
				 "         GOODSIG %s  %s\n"\
				 "         FP_KEY  %s"),
#endif
			       sh.user, 
			       file_path('C', 'R'), gp.conf_id, gp.conf_fp);
	      return (0);
	    }
	}
      else
	{
	  status = SH_GPG_BADSIGN;
	}
#else
      if ( (both < 2) || 
	   ((sl_strcmp(gp.data_id, gp.conf_id) == 0) &&
	    (sl_strcmp(gp.data_fp, gp.conf_fp) == 0))
	   )
	sh_error_handle (SH_ERR_FATAL,
#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
			 _("STARTUP yule    -- user %s\n"\
			   "         GOODSIG %s %s\n"\
			   "         FP_KEY  %s"),
#else
			 _("STARTUP samhain -- user %s\n"\
			   "         GOODSIG %s %s\n"\
			   "         FP_KEY  %s"),
#endif
			 sh.user, 
                         file_path('C', 'R'), gp.conf_id, gp.conf_fp);
      else
	sh_error_handle (SH_ERR_FATAL,
#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
			 _("STARTUP yule    -- user %s\n"\
			   "         GOODSIG %s %s\n"\
			   "         FP_KEY  %s\n"\
			   "         GOODSIG_DAT %s %s\n"\
			   "         FP_KEY_DAT  %s"),
#else
			 _("STARTUP samhain -- user %s\n"\
			   "         GOODSIG %s %s\n"\
			   "         FP_KEY  %s\n"\
			   "         GOODSIG_DAT %s  %s\n"\
			   "         FP_KEY_DAT  %s"),
#endif
			 sh.user, 
                         file_path('C', 'R'), gp.conf_id, gp.conf_fp,
			 file_path('D', 'R'), gp.data_id, gp.data_fp);
        return (0);
#endif
    }
  if (SH_GPG_BAD == status)
    sh_error_handle (SH_ERR_FATAL,
#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
		     _("STARTUP yule    -- user %s\n"\
		       "         ERRSIG: PGP error (access, "\
		       "trust, or checksum)"),
#else
                     _("STARTUP samhain -- user %s\n"\
		       "         ERRSIG: PGP error (access, "\
		       "trust, or checksum)"),
#endif
		     sh.user);
  if (SH_GPG_BADSIGN == status)
    sh_error_handle (SH_ERR_FATAL,
#if defined(SH_WITH_SERVER) && !defined(SH_WITH_CLIENT)
		     _("STARTUP yule    -- user %s\n"\
		       "         BADSIG: Configuration/database."),
#else
                     _("STARTUP samhain -- user %s\n"\
		       "         BADSIG: Configuration/database."),
#endif
		     sh.user);
  return (-1);
}  

/* #ifdef WITH_GPG */
#endif








