/* SAMHAIN file system integrity testing                                   */
/* Copyright (C) 1999 Rainer Wichmann                                      */
/*                                                                         */
/*  This program is free software; you can redistribute it                 */
/*  and/or modify                                                          */
/*  it under the terms of the GNU General Public License as                */
/*  published by                                                           */
/*  the Free Software Foundation; either version 2 of the License, or      */
/*  (at your option) any later version.                                    */
/*                                                                         */
/*  This program is distributed in the hope that it will be useful,        */
/*  but WITHOUT ANY WARRANTY; without even the implied warranty of         */
/*  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          */
/*  GNU General Public License for more details.                           */
/*                                                                         */
/*  You should have received a copy of the GNU General Public License      */
/*  along with this program; if not, write to the Free Software            */
/*  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.              */

#include <stdlib.h>
#include <string.h>
#include <stdio.h>
#include <sys/stat.h>
#include <unistd.h>

#include "config_xor.h"

#ifndef HAVE_LSTAT
#define lstat(a,b) stat(a,b)
#endif

#ifdef HAVE_MEMORY_H
#include <memory.h>
#endif

#if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE)

#include "sh_hash.h"
#include "sh_utils.h"
#include "sh_error.h"
#include "sh_tiger.h"

#if defined(SH_WITH_CLIENT)
#include "sh_forward.h"
#endif

/* MAX_PATH_STORE must be >= KEY_LEN
 */
#define MAX_PATH_STORE 254

typedef struct store_info {
  unsigned short   mark;
  char             fullpath[MAX_PATH_STORE+2];
  char             linkpath[MAX_PATH_STORE+2];
  unsigned int     mode;
  unsigned int     linkmode;
  char             c_mode[11];
  long             owner;
  char             c_owner[OWNER_MAX+2];
  long             group;
  char             c_group[GROUP_MAX+2];
  unsigned long    hardlinks;
  unsigned long    ino;
  unsigned long    size;
  unsigned long    atime;
  unsigned long    mtime;
  unsigned long    ctime;
  char             checksum[KEY_LEN+1];
} sh_filestore_t;
  
typedef struct file_info {
  sh_filestore_t   theFile;
  int              visited;
  struct           file_info * next;
} sh_file_t;

  static char  *policy[] = {
    N_("[]"),
    N_("[ReadOnly]"),
    N_("[LogFiles]"),
    N_("[GrowingLogs]"),
    N_("[IgnoreNone]"),
    N_("[IgnoreAll]"),
    N_("[Attributes]"),
    NULL
  };


static char * pM = N_("POLICY MISSING %s\n");
static char * pA = N_("POLICY ADDED   %s\n");
static char * pG = N_("POLICY %s %s\n");
static char * h1 = N_("            inode ln       mode    owner    group       size                 mtime                 %-5s\n");
static char * h2 = N_("          ------- -- ---------- -------- -------- ---------- --------------------- ---------------------\n");
static char * fLm = N_("        < %7ld %2d %10s %8s %8s %10ld %21s %21s %s");
static char * fNm = N_("        < %7ld %2d %10s %8s %8s %10ld %21s %21s");
static char * fLa = N_("        > %7ld %2d %10s %8s %8s %10ld %21s %21s %s");
static char * fNa = N_("        > %7ld %2d %10s %8s %8s %10ld %21s %21s");

static
char * make_fmt (char * poli, char * head1, char * head2, 
		 char * form, char * form2)
{
  int len;
  char * p;

  len = sl_strlen(poli) + sl_strlen(head1) + sl_strlen(head2) + sl_strlen(form) + 1;
  if (form2 != NULL)
    len += (sl_strlen(form2) + 2);

  p   = SH_ALLOC(len);

  sl_strlcpy(p, _(poli),  len);
  sl_strlcat(p, _(head1), len);
  sl_strlcat(p, _(head2), len);
  sl_strlcat(p, _(form),  len);
  if (form2 != NULL)
    {
      sl_strlcat(p, "\n", len);
      sl_strlcat(p, _(form2),  len);
    }

  return (p);
}


/**********************************
 *
 * hash table functions
 *
 **********************************
 */

#include "sh_hash.h"

/* must fit an int              */
#define TABSIZE 2048  

/* must fit an unsigned short   */
/* changed for V0.8, as the     */
/* database format has changed  */
#define REC_MAGIC 18

static sh_file_t * tab[TABSIZE];

/* compute hash function
 */
static int hashfunc(char *s) {

  unsigned n = 0; 

  for ( ; *s; s++) 
    n = 31 * n + *s; 
  return n % TABSIZE; 
} 

/* search for files not visited, and check whether they exist
 */
static void hash_unvisited (sh_file_t *p, ShErrLevel level)
{
  struct stat buf;
  int i;
  char * tmp;
  char * tmp_lnk;
  char timstr2c[32];
  char timstr2m[32];
  char * fmt;

  if (p->next != NULL)
    hash_unvisited (p->next, level);

  if (p->visited == BAD)
    {
      sl_set_suid(); 
      i = lstat(p->theFile.fullpath, &buf);
      sl_unset_suid();

      if (0 != i)
	{
	  sl_strlcpy (timstr2c, sh_unix_time (p->theFile.ctime),   32);
	  sl_strlcpy (timstr2m, sh_unix_time (p->theFile.mtime),   32);

	  tmp = sh_util_safe_name(p->theFile.fullpath);
	  if (p->theFile.c_mode[0] == 'l') 
	    {    
	      tmp_lnk     = sh_util_safe_name(p->theFile.linkpath);
	      fmt = make_fmt (pM, h1, h2, fLm, NULL);
	      sh_error_handle (level, 
			       fmt,
			       tmp,
			       _("ctime"),
			       p->theFile.ino,
			       p->theFile.hardlinks,
			       p->theFile.c_mode,
			       p->theFile.c_owner,
			       p->theFile.c_group,
			       p->theFile.size,
			       timstr2m,
			       timstr2c,
			       tmp_lnk);
	      SH_FREE(tmp_lnk);
	      SH_FREE(fmt);
	    }
	  else
	    {    
	      fmt = make_fmt (pM, h1, h2, fNm, NULL);
	      sh_error_handle (level,
			       fmt,
			       tmp,
			       _("ctime"),
			       p->theFile.ino,
			       p->theFile.hardlinks,
			       p->theFile.c_mode,
			       p->theFile.c_owner,
			       p->theFile.c_group,
			       p->theFile.size,
			       timstr2m,
			       timstr2c);
	      SH_FREE(fmt);
	    }
	  SH_FREE(tmp);
	}
    }

  p->visited = BAD;
  return;
}
void sh_hash_unvisited (ShErrLevel level)
{
  int i;

  for (i = 0; i < TABSIZE; ++i)
    {
      if (tab[i] != NULL) hash_unvisited (tab[i], level);
    }
  return;
}

/* delete hash array
 */
static void hash_kill (sh_file_t *p)
{

  if (p->next != NULL)
    hash_kill (p->next);

  SH_FREE(p);
  return;
}

/* get info out of hash array
 */
static sh_file_t * hashsearch (char * s) {

  sh_file_t * p;

  for (p = tab[hashfunc(s)]; p; p = p->next) 
    if (sl_strcmp(s, p->theFile.fullpath) == 0) 
      return p; 
  return NULL;
} 

/* insert into hash array
 */
static void hashinsert (sh_file_t * s) {

  sh_file_t * p;

  int key = hashfunc(s->theFile.fullpath);

  if (tab[key] == NULL) {
    tab[key] = s;
    tab[key]->next = NULL;
    return;
  } else {
    p = tab[key];
    while (1) {
      if (p->next == NULL) {
	p->next = s;
	p->next->next = NULL;
	return;
      }
      p = p->next;
    }
  }
  /* notreached */
}


/* check functions
 */

static int IsInit = 0;

void sh_hash_init ()
{
  sh_file_t * p;
  sh_filestore_t ft;
  static SL_TICKET fd = (-1);
  long i;
  int count = 0;
  char * line;

  if (IsInit == 1) 
    return;

  for (i = 0; i < TABSIZE; ++i) tab[i] = NULL;

#if defined(SH_WITH_CLIENT)
  if (fd == (-1) && 0 == sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
    {
      sh_error_handle (SH_ERR_INFO,
		       _("Downloading database file"));
      fd = sh_forward_req_file(_("DATA"));
      if (SL_ISERROR(fd))
	exit (EXIT_FAILURE);

      sl_rewind (fd);

      tiger_fd = fd;
      sl_strlcpy (sh.dat_hash, 
		  sh_tiger_hash (file_path('C', 'R'), 
				 TIGER_FD, 0),
		  KEY_LEN+1);
    }
  else 
#endif
    if (fd == (-1))
      {
	if ( SL_ISERROR(fd = sl_open_read(file_path('D', 'R'), SL_YESPRIV))) {
	  sh_error_handle (SH_ERR_FATAL,
			   _("PANIC -- no database found."));
	  exit (EXIT_FAILURE);
	}
	
	tiger_fd = fd;
	if (0 != sl_strncmp(sh.dat_hash, 
			 sh_tiger_hash (file_path('C', 'R'), 
					TIGER_FD, 0),
			 KEY_LEN)) {
	  sh_error_handle (SH_ERR_FATAL,
			   _("PANIC -- database modified "\
			     "since program start."));
	  exit (EXIT_FAILURE);
	}
      }


  sl_rewind (fd);

  /* fast forward to start of data
   */
  line = SH_ALLOC(SH_BUFSIZE+1);
  while (1) {
    i =  sh_unix_getline (fd, line, SH_BUFSIZE);
    if (i < 0 ) {
      SH_FREE(line);
      sh_error_handle (SH_ERR_FATAL,
		       _("PANIC -- database no start marker found."));
      exit (EXIT_FAILURE);
    }
#if defined(SH_STEALTH)
    if (0 == sl_strncmp (line, N_("[SOF]"), 5)) 
#else
    if (0 == sl_strncmp (line, _("[SOF]"),  5)) 
#endif
      break;
  }
  SH_FREE(line);

  while (1) 
    {
      p = SH_ALLOC(sizeof(sh_file_t));
      ++count;
      i = sl_read(fd, &ft, sizeof(sh_filestore_t));
      if ( SL_ISERROR(i) || i == 0) 
	{
	  SH_FREE(p);
	  break;
	}
      
      if (ft.mark != REC_MAGIC)
	{
	  SH_FREE(p);
	  break;
	}
      
#if defined(SH_STEALTH)
      sh_do_decode(ft.fullpath, sl_strlen(ft.fullpath));

      sh_do_decode(ft.c_mode,   sl_strlen(ft.c_mode));
      sh_do_decode(ft.c_owner,  sl_strlen(ft.c_owner));
      sh_do_decode(ft.c_group,  sl_strlen(ft.c_group));
      sh_do_decode(ft.checksum, sl_strlen(ft.checksum));


      if (ft.c_mode[0] == 'l') 
	{  
	  sh_do_decode(ft.linkpath, sl_strlen(ft.linkpath));
	}
#endif

      memcpy( &(*p).theFile, &ft, sizeof(sh_filestore_t) );
      p->visited = BAD;
      
      hashinsert (p); 
      ft.mark = (REC_MAGIC + 5);
    }

#if defined(SH_WITH_CLIENT)
  if (0 == sl_strcmp(file_path('D', 'R'), _("REQ_FROM_SERVER")))
    {
      sl_rewind (fd);
    }
  else
    {
      sl_close (fd);
      fd = -1;
    }
#else
  sl_close (fd);
  fd = -1;
#endif

  IsInit = 1;
  return;
}
  
/* delete hash array
 */
void sh_hash_hashdelete ()
{
  int i;

  if (IsInit == 0) return;

  for (i = 0; i < TABSIZE; ++i) 
    if (tab[i] != NULL) hash_kill (tab[i]);

  IsInit = 0;
  return;
}

  
void sh_hash_pushdata (file_type * buf, char * fileHash)
{
  SL_TICKET fd;
  sh_filestore_t p;
  static int isfirst = 1;

  if (0 == sl_strcmp(file_path('D', 'W'), _("REQ_FROM_SERVER")))
    return;

  if ( SL_ISERROR(fd = sl_open_write(file_path('D', 'W'), SL_YESPRIV)))
    return;

  if ( SL_ISERROR(sl_forward(fd)))
    return;

  if (sl_strlen(buf->fullpath) <= MAX_PATH_STORE) {
    sl_strlcpy(p.fullpath, buf->fullpath, MAX_PATH_STORE+1);
    if (sl_strlen(buf->fullpath) < (MAX_PATH_STORE-3))
      {
	p.fullpath[MAX_PATH_STORE-2] = '\0';
	p.fullpath[MAX_PATH_STORE-1] = '\n';
      }
  } else {
    sl_strlcpy(p.fullpath, 
	       sh_tiger_hash (buf->fullpath,
			      TIGER_DATA,sl_strlen(buf->fullpath)), 
	       KEY_LEN+1);
  }

#if defined(SH_STEALTH)
  sh_do_encode(p.fullpath, sl_strlen(p.fullpath));
#endif

  if (buf->c_mode[0] == 'l') {  
    if (sl_strlen(buf->linkpath) <= MAX_PATH_STORE) {
      sl_strlcpy(p.linkpath, buf->linkpath, MAX_PATH_STORE+1);  
    } else {
      sl_strlcpy(p.linkpath, 
		 sh_tiger_hash (buf->linkpath,
				TIGER_DATA,sl_strlen(buf->linkpath)),
		 KEY_LEN+1);
    }

#if defined(SH_STEALTH)
    sh_do_encode(p.linkpath, sl_strlen(p.linkpath));
#endif
  }


  p.mark = REC_MAGIC;
  sl_strlcpy(p.c_mode,   buf->c_mode,   11);
  sl_strlcpy(p.c_group,  buf->c_group,  GROUP_MAX+1);
  sl_strlcpy(p.c_owner,  buf->c_owner,  OWNER_MAX+1);
  sl_strlcpy(p.checksum, fileHash,      KEY_LEN+1);

#if defined(SH_STEALTH)
  sh_do_encode(p.c_mode,   sl_strlen(p.c_mode));
  sh_do_encode(p.c_owner,  sl_strlen(p.c_owner));
  sh_do_encode(p.c_group,  sl_strlen(p.c_group));
  sh_do_encode(p.checksum, sl_strlen(p.checksum));
#endif


  p.linkmode    = buf->linkmode;
  p.hardlinks   = buf->hardlinks;
  p.mode  = buf->mode;
  p.ino   = buf->ino;
  p.size  = buf->size;
  p.mtime = buf->mtime;
  p.atime = buf->atime;
  p.ctime = buf->ctime;
  p.owner = buf->owner;
  p.group = buf->group;

  /* write the start marker 
   */
  if (isfirst == 1) {
#if defined(SH_STEALTH)
    sl_write      (fd,        "\n", 1);
    sl_write_line (fd, N_("[SOF]"), 5);
#else
    sl_write_line (fd, _("\n[SOF]"),  6);
#endif
    isfirst = 0;
  }
      
  sl_write (fd, &p, sizeof(sh_filestore_t));

  sl_close (fd);

  return;
}

int sh_hash_have_it (char * newname)
{
  sh_file_t * p;

  if (IsInit != 1) sh_hash_init();

  if (sl_strlen(newname) <= MAX_PATH_STORE) {
    p = hashsearch(newname);
  } else {
    p = hashsearch ( sh_tiger_hash(newname, TIGER_DATA, sl_strlen(newname)) );
  }
  if (p == NULL) 
    return (-1);
  return (0);
}

#define MODI_CHK (1 << 0)
#define MODI_LNK (1 << 1)
#define MODI_INO (1 << 2)
#define MODI_USR (1 << 3)
#define MODI_GRP (1 << 4)
#define MODI_MTM (1 << 5)
#define MODI_CTM (1 << 6)
#define MODI_ATM (1 << 7)
#define MODI_SIZ (1 << 8)
#define MODI_MOD (1 << 9)
#define MODI_HLN (1 << 10)
  
void sh_hash_compdata (int class,
		  file_type * theFile, char * fileHash)
{
  sh_file_t * p;
  char * tmp;
  char * fmt;
  char * tmp_lnk;
  char * tmp_lnk_old;
  char timstr1c[32];
  char timstr2c[32];
  char timstr1a[32];
  char timstr2a[32];
  char timstr1m[32];
  char timstr2m[32];
  char linkHash[KEY_LEN+1];
  int  maxcomp;

  unsigned long modi_mask = 0;

  if (IsInit != 1) sh_hash_init();


  /* --------  find the entry for the file ----------------       */

  if (sl_strlen(theFile->fullpath) <= MAX_PATH_STORE) {
    p = hashsearch(theFile->fullpath);
  } else {
    p = hashsearch( sh_tiger_hash(theFile->fullpath, 
				  TIGER_DATA, 
				  sl_strlen(theFile->fullpath))
		    );
  }
  if (p == NULL) {
    sl_strlcpy (timstr2c, sh_unix_time (theFile->ctime),   32);
    sl_strlcpy (timstr2a, sh_unix_time (theFile->atime),   32);
    sl_strlcpy (timstr2m, sh_unix_time (theFile->mtime),   32);

    tmp = sh_util_safe_name(theFile->fullpath);
    if (theFile->c_mode[0] == 'l') 
      {    
	fmt = make_fmt (pA, h1, h2, fLa, NULL);
	tmp_lnk     = sh_util_safe_name(theFile->linkpath);
	sh_error_handle (ShDFLevel[class],
			 fmt,
			 tmp,
			 (class == SH_LEVEL_NOIGNORE) ? 
			 _("atime") : _("ctime"),
			 theFile->ino,
			 theFile->hardlinks,
			 theFile->c_mode,
			 theFile->c_owner,
			 theFile->c_group,
			 theFile->size,
			 timstr2m,
			 (class == SH_LEVEL_NOIGNORE) ? 
			 timstr2a : 
			 timstr2c,
			 tmp_lnk);
	SH_FREE(tmp_lnk);
	SH_FREE(fmt);
      }
    else
      {    
	fmt = make_fmt (pA, h1, h2, fNa, NULL);
	sh_error_handle (ShDFLevel[class],
			 fmt,
			 tmp,
			 (class == SH_LEVEL_NOIGNORE) ? 
			 _("atime") : _("ctime"),
			 theFile->ino,
			 theFile->hardlinks,
			 theFile->c_mode,
			 theFile->c_owner,
			 theFile->c_group,
			 theFile->size,
			 timstr2m,
			 (class == SH_LEVEL_NOIGNORE) ? 
			 timstr2a : 
			 timstr2c);
	SH_FREE(fmt);
      }
    SH_FREE(tmp);
    return;
  }
#ifdef MEM_DEBUG
  sh_error_handle (SH_ERR_ALL,
		   _("  tiger old: %s\n   tiger new: %s"), 
		   fileHash, 
		   p->theFile.checksum);
#endif


  if ( 
      (sl_strncmp (fileHash, p->theFile.checksum, 50) != 0) && 
      ( class != SH_LEVEL_LOGFILES    &&
	class != SH_LEVEL_LOGGROW     &&
	class != SH_LEVEL_ATTRIBUTES  &&
	class != SH_LEVEL_ALLIGNORE 
      )
     ) {
    modi_mask |= MODI_CHK;
  }

  if (p->theFile.c_mode[0] == 'l') {

    if (sl_strlen(theFile->linkpath) <= MAX_PATH_STORE) {
      sl_strlcpy(linkHash, 
		 sh_tiger_hash(theFile->linkpath, 
			       TIGER_DATA,
			       sl_strlen(theFile->linkpath)), 
		 MAX_PATH_STORE+1);
      maxcomp = MAX_PATH_STORE;
    } else {
      sl_strlcpy(linkHash, theFile->linkpath, KEY_LEN + 1);
      maxcomp = KEY_LEN;
    }

    if ( 
	(sl_strncmp (linkHash, p->theFile.linkpath, maxcomp) != 0) &&
	( class != SH_LEVEL_ALLIGNORE 
	  && class != SH_LEVEL_ATTRIBUTES
	  )
	) {
      modi_mask |= MODI_LNK;
    }
  }

    
  if ( (theFile->ino != p->theFile.ino ) &&
      ( class != SH_LEVEL_ALLIGNORE &&  
	class != SH_LEVEL_ATTRIBUTES
      )
     ) {
    modi_mask |= MODI_LNK;
  }
    
  if ( (theFile->hardlinks != p->theFile.hardlinks ) &&
      ( class != SH_LEVEL_ALLIGNORE &&  
	class != SH_LEVEL_ATTRIBUTES
      )
     ) {
    modi_mask |= MODI_HLN;
  }
    
  if ( (theFile->mode != p->theFile.mode) &&
      ( class != SH_LEVEL_ALLIGNORE 
      )
     ) {
    modi_mask |= MODI_MOD;
  }

  if ( (theFile->owner != p->theFile.owner) &&
      ( class != SH_LEVEL_ALLIGNORE 
      )
     ) {
    modi_mask |= MODI_USR;
  }

  if ( (theFile->group != p->theFile.group) &&
      ( class != SH_LEVEL_ALLIGNORE 
      )
    ) {
    modi_mask |= MODI_GRP;
  }

  
  if ( (theFile->mtime != p->theFile.mtime) &&
      ( class != SH_LEVEL_LOGFILES  &&
	class != SH_LEVEL_LOGGROW   &&
	class != SH_LEVEL_ALLIGNORE && 
	class != SH_LEVEL_ATTRIBUTES
      )
    ) {
    modi_mask |= MODI_MTM;
  }
  
  if ( (theFile->atime != p->theFile.atime) &&
      ( class == SH_LEVEL_NOIGNORE 
      )
    ) {
    modi_mask |= MODI_ATM;
  }
  
  /* Resetting the access time will set a new ctime. Thus, either we ignore
   * the access time or the ctime for NOIGNORE
   */
  if ( (theFile->ctime != p->theFile.ctime) &&
       ( class != SH_LEVEL_LOGFILES  &&
	 class != SH_LEVEL_LOGGROW   &&
	 class != SH_LEVEL_ALLIGNORE &&
	 class != SH_LEVEL_NOIGNORE  &&
	 class != SH_LEVEL_ATTRIBUTES
	 )
       ) {
    modi_mask |= MODI_CTM;
  }
   
  
  if ( (theFile->size != p->theFile.size) &&
      ( class != SH_LEVEL_ALLIGNORE &&
        class != SH_LEVEL_LOGFILES  &&
	class != SH_LEVEL_LOGGROW   &&
	class != SH_LEVEL_ATTRIBUTES   
      )
     ) {
    modi_mask |= MODI_SIZ;
  }
  
  if ( (theFile->size < p->theFile.size) &&
      ( class == SH_LEVEL_LOGGROW
      )
     ) {
    modi_mask |= MODI_SIZ;
  }


  /* report on modified files
   */
  if (modi_mask != 0)
    {    
      sl_strlcpy (timstr1c, sh_unix_time (p->theFile.ctime), 32);
      sl_strlcpy (timstr2c, sh_unix_time (theFile->ctime),   32);
      sl_strlcpy (timstr1a, sh_unix_time (p->theFile.atime), 32);
      sl_strlcpy (timstr2a, sh_unix_time (theFile->atime),   32);
      sl_strlcpy (timstr1m, sh_unix_time (p->theFile.mtime), 32);
      sl_strlcpy (timstr2m, sh_unix_time (theFile->mtime),   32);

      tmp = sh_util_safe_name(theFile->fullpath);
      if (theFile->c_mode[0] == 'l') 
	{    
	  tmp_lnk     = sh_util_safe_name(theFile->linkpath);
	  tmp_lnk_old = sh_util_safe_name(p->theFile.linkpath);
	  fmt = make_fmt (pG, h1, h2, fLm, fLa);
	  sh_error_handle (ShDFLevel[class], 
			   fmt,
			   _(policy[class]), tmp,
			   (class == SH_LEVEL_NOIGNORE) ? 
			   _("atime") : _("ctime"),
			   p->theFile.ino,
			   p->theFile.hardlinks,
			   p->theFile.c_mode,
			   p->theFile.c_owner,
			   p->theFile.c_group,
			   p->theFile.size,
			   timstr1m,
			   (class == SH_LEVEL_NOIGNORE) ? 
			   timstr1a : 
			   timstr1c,
			   tmp_lnk_old,
			   theFile->ino,
			   theFile->hardlinks,
			   theFile->c_mode,
			   theFile->c_owner,
			   theFile->c_group,
			   theFile->size,
			   timstr2m,
			   (class == SH_LEVEL_NOIGNORE) ? 
			   timstr2a : 
			   timstr2c,
			   tmp_lnk);
	  SH_FREE(fmt);
	  SH_FREE(tmp_lnk);
	  SH_FREE(tmp_lnk_old);
	}
      else
	{    
	  fmt = make_fmt (pG, h1, h2, fNm, fNa);
	  sh_error_handle (ShDFLevel[class], 
			   fmt,
			   _(policy[class]), tmp,
			   (class == SH_LEVEL_NOIGNORE) ? 
			   _("atime") : _("ctime"),

			   p->theFile.ino,
			   p->theFile.hardlinks,
			   p->theFile.c_mode,
			   p->theFile.c_owner,
			   p->theFile.c_group,
			   p->theFile.size,
			   timstr1m,
			   (class == SH_LEVEL_NOIGNORE) ? 
			   timstr1a : 
			   timstr1c,

			   theFile->ino,
			   theFile->hardlinks,
			   theFile->c_mode,
			   theFile->c_owner,
			   theFile->c_group,
			   theFile->size,
			   timstr2m,
			   (class == SH_LEVEL_NOIGNORE) ? 
			   timstr2a : 
			   timstr2c);
	  SH_FREE(fmt);
	}
      SH_FREE(tmp);
    }

  p->visited = GOOD;

  return;
}
  


/* if defined(SH_WITH_CLIENT) || defined(SH_STANDALONE) */

#endif
